/*
 * Copyright (C) 1995  Sam Moore, Warren Jasper
 * Copyright (C) 1995  Ricky Beam, Rob Cox
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

/***************************************************************************
 *
 * a2dc.c
 *
 ***************************************************************************/

#include <linux/autoconf.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/timer.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/signal.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <stdlib.h>
#include "adc.h"
#include "a2dc.h"

/***************************************************************************
 *
 * Prototype of pubic and private functions.
 *
 ***************************************************************************/

static int  SoftTrigRead(ChanRec *chan);
static int  PacerRead(ChanRec *chan);
static void SetGains(BYTE gain);
static int  SetPacerFreq(BoardRec *board);
static void StopPacer();
static void SetChannelMux(BYTE lowChan, BYTE highChan);
static void LoadPacer(BoardRec *board);
static void LoadCounter0( LONG arg );

static int das_read(struct inode *iNode, struct file *filePtr,
		    char *buf, int count);
static int das_write(struct inode *iNode, struct file *filePtr,
		     char *buf, int count);
static int das_open(struct inode *iNode, struct file *filePtr);
static void das_close(struct inode *iNode, struct file *filePtr);
static int das_ioctl(struct inode *iNode, struct file *filePtr,
		     unsigned int cmd, LONG arg);
static void das0_ReadInterrupt();
static void das0_TimerHandler(LONG);
static void doDMA(int samples);

int init_module(void);
void cleanup_module(void);

/***************************************************************************
 *
 * Global data.
 *
 ***************************************************************************/

static int WordsToRead;                     /* number of conversions until done  */
static int ISR_WordsToRead;                 /* number of conversions until done  */
static ChanRec *CurrChan = NULL;            /* pointer to Chan[minor]            */
static int MajorNumber = DEFAULT_MAJOR_DEV; /* Major number compiled in          */
/* static ChanRec Chan[AD_CHANNELS];           /* Channel specific information      */
static ChanRec Chan[32];                    /* Channel specific information      */
static BoardRec BoardData;                  /* Board specific information        */

static WORD *KernBuff;                      /* Pointer to User space             */
static WORD *DMAbuf[2];                     /* static DMA buffer space           */
static BYTE DMAptr = 0;                     /* pointer to the active DMAbuf      */

char kernel_version[] = UTS_RELEASE;        /* Required to by module             */
struct wait_queue  *das0_wait;              /* wait semaphore                    */
struct timer_list TimerList = {NULL, NULL, 0, 0, das0_TimerHandler};

/***************************************************************************
 *
 * read() service function
 *
 ***************************************************************************/

static int das_read(struct inode *iNode, struct file *filePtr, char *buf, int count)
{
  int	  stat;
  BYTE  value;
  int   minor = MINOR(iNode->i_rdev);
  
#ifdef DEBUG
  printk("das_read(): Entering das_read()\n");
#endif
  
  if ( minor >= 0 && minor < AD_CHANNELS ) {
    
    /* Check that data can be written to file. */
    
    if ((stat = verify_area(VERIFY_WRITE, buf,
			    sizeof(WORD) * count)) != 0) {
      printk("das_read: Failed VERIFY_WRITE\n");
      return -1;
    }
    
    /* Area is secure... */
    KernBuff = (WORD *) buf;
    
    /* Set modes for this channel */
    
    SetGains((BYTE)Chan[minor].gain);
    Chan[minor].count = count;
    SetChannelMux(minor, Chan[minor].highScan);
    
    /* Read */
    
    switch (Chan[minor].mode) {
    case ADC_SOFT_TRIGGER:
#ifdef DEBUG
      printk("das_read(): Entering ADC_SOFT_TRIGGER mode.\n");
#endif
      if (SoftTrigRead(&Chan[minor])) {
	printk("das_read: SoftTrigRead() failed.\n");
	return(-1);
      } 
      break;
      
    case ADC_EXTERNAL_TRIGGER:
#ifdef DEBUG
      printk("das_read(): Entering ADC_EXTERNAL_TRIGGER mode.\n");
#endif
      if (PacerRead(&Chan[minor])) {
	printk("das_read: ExternalRead() failed.\n");
	return(-1);
      }
      break;
      
    case ADC_PACER_CLOCK:
#ifdef DEBUG
      printk("das_read(): Entering ADC_PACER_CLOCK mode.\n");
#endif
      StopPacer();		     /* disable pacer if in pacer mode */
      outb_p(0x40, CONVERT_DIS_REG); /* Disable Conversions            */
      if (PacerRead(&Chan[minor])) {
	printk("das_read: PacerRead() failed.\n");
	return(-1);
      } 
      break;
      
    case ADC_PACER_EXTERN_TRIG:
#ifdef DEBUG
      printk("das_read(): Entering ADC_PACER_EXTERN_TRIG mode.\n");
#endif
      StopPacer();                   /* disable pacer if in pacer mode */
      outb_p(0x40, CONVERT_DIS_REG); /* Disable Conversions            */
      if (PacerRead(&Chan[minor])) {
	printk("das_read: PacerRead() with external tigger failed.\n");
	return(-1);
      } 
      break;
    }
    
    return(Chan[minor].count);	/* return number of samples read */
    
  } else if (minor <= (AD_CHANNELS + DIO_PORTS)) {
    
    /* check to see if reading a value from the DIO */
    
    register int port = (minor&0x7)?DIO_REG_EXTENDED + (minor&0x7) - 1:DIO_REG;
    
    value = inb(port);
    cli();
    put_fs_byte(value, buf);
    sti();
    return 1;
  }
  
  printk("das_read: Incorrect minor number (%d).\n", minor);
  return -1;
}

/***************************************************************************
 *
 * write() service function
 *
 ***************************************************************************/

static int das_write(struct inode *iNode, struct file *filePtr, char *buf, int count)
{
  int  minor = MINOR(iNode->i_rdev);
  BYTE value;
  WORD voltage;
  
  if (minor >= 0 && minor < AD_CHANNELS) {
    
    /* Attempting to WRITE to a analog->digital port */
    
    return(-EINVAL);
    
  } else if (minor <= (AD_CHANNELS + DIO_PORTS)) {
    
    register int port =
      (minor&0xf)?DIO_REG_EXTENDED + (minor&0xf) - 1:DIO_REG;
    register int mask = (minor&0xf)?0xff:0xf;
    
    cli();
    value = get_fs_byte(buf) & mask;
    outb_p( value, port);
    sti();
    
#ifdef DEBUG
    printk("das_write(dio): DIO Port %#x set to %#x\n",
	   (minor&0xf), value & mask);
#endif
    
    return 1;
    
  } else if (minor < (DA_MINOR + DA_PORTS)) {
    
    register int port = minor - DA_MINOR + DA_PORT0;
    
    cli();
    voltage = get_fs_word(buf);
    outw_p(voltage, port);
    sti(); 
    
#ifdef DEBUG
    printk("das_write(dac): D2A Channel %#x set to %#x\n",
	   port - DA_PORT0, voltage);
#endif
    
    return(2);
    
  } else {
    
    printk("das_write(): bad minor number = %d\n", minor );
    
    return -1;
  }
}


/***************************************************************************
 *
 * open() service handler
 *
 ***************************************************************************/

static int das_open(struct inode *iNode, struct file *filePtr)
{
  int minor = MINOR(iNode->i_rdev);
  
  /* 
     check if device is already open: only one process may read from a
     port at a time.  There is still the possibility of two processes 
     reading from two different channels messing things up. However,
     the overhead to check for this may not be worth it.
     */
  
  if ( Chan[minor].open == TRUE ) {
    return -EBUSY;
  }
  
  MOD_INC_USE_COUNT;
  
  Chan[minor].open = TRUE;	/* The device is open */
  Chan[minor].gain = BP_10_00V;	/* +/- 10V */
  Chan[minor].mode = filePtr->f_flags; /* set trigger mode */
  Chan[minor].highScan = (BYTE) minor;
  Chan[minor].lowScan = (BYTE) minor;
  
  if (Chan[minor].mode == ADC_PACER_EXTERN_TRIG) {
    BoardData.pacerReg = 0x1;
  } else {
    BoardData.pacerReg = 0x0;
  }
  
  outb_p(0x00, STATUS_REG);	/* Clear INT bit */
  outb_p(0x0, MISC_REG);	/* Put in compatibility mode */
  
#ifdef DEBUG
  printk("das_open(): minor %d mode %d.\n", minor, Chan[minor].mode);
#endif
  
  return 0;   
}

/***************************************************************************
 *
 * close() service handler
 *
 ***************************************************************************/

static void das_close(struct inode *iNode, struct file *filePtr)
{
  int minor = MINOR(iNode->i_rdev);
  
#ifdef DEBUG
  printk("%s: close() of minor number %d.\n", ADAPTER_ID, minor);
#endif
  
  Chan[minor].open = FALSE;
  MOD_DEC_USE_COUNT;
}

/***************************************************************************
 *
 * iotctl() service handler
 *
 ***************************************************************************/

/* Note:
   Remember that FIOCLEX, FIONCLEX, FIONBIO, and FIOASYN are reserved ioctl cmd numbers
   */

static int das_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg)
{
  int minor = MINOR(iNode->i_rdev);
  BYTE bReg;
  
  switch (cmd) {
  case ADC_SET_GAINS:
    Chan[minor].gain = (BYTE)arg;
#ifdef DEBUG
    printk("%s: ioctl(ADC_SET_GAINS) %#x\n", ADAPTER_ID, (int) arg);
#endif
    break;
    
  case ADC_SET_PACER_FREQ:
    if ( BoardData.freq > MAX_FREQ  && Chan[minor].mode == ADC_PACER_CLOCK) {
      printk("ioctl:  BoardData.freq = %ld out of range.\n",
	     BoardData.freq);
      return -1;
    } else {
      BoardData.freq = (LONG)arg;
      SetPacerFreq(&BoardData);
      LoadPacer(&BoardData);	/* load the board frequency now */
      outb_p(BoardData.pacerReg, PACER_CLOCK_REG);  
    }
    break;
    
  case ADC_GET_PACER_FREQ:
    cli();
    put_fs_long(BoardData.freq, (long*) arg);
    sti();
    break;
    
  case ADC_STOP_PACER:
    StopPacer();
    break;
    
  case CTR0:
    if ( arg == 1 ) {
      BoardData.pacerReg |= 0x2;
    } else {
      BoardData.pacerReg &= ~(0x2);
    }
    outb_p(BoardData.pacerReg, PACER_CLOCK_REG);  
    break;
    
  case COUNTER0:
    LoadCounter0( (LONG) arg );
    break;
    
  case ADC_DIO_PRESET:
    if ( arg == -1 ) {
      Chan[minor].adc_dio_pretrig = FALSE;
    } else {
      Chan[minor].adc_dio_pretrig = TRUE;
      Chan[minor].dio_value = (BYTE) (arg & 0xf);
    }
    
#ifdef DEBUG
    printk("ioctl(ADC_DIO_PRESET):  value = %#x\n", Chan[minor].dio_value);
#endif
    
    break;
    
  case ADC_SETMUX_LIMITS:
    if (minor > AD_CHANNELS)
      return(-EINVAL);
    Chan[minor].highScan = minor + (arg & 0xf);
    if (Chan[minor].highScan > AD_CHANNELS) {
      Chan[minor].highScan = minor;
      BoardData.pacerReg &= 0x3;
      return(-EINVAL);
    }
    BoardData.pacerReg |= ((arg & 0xf) << 4);
    outb_p(BoardData.pacerReg, PACER_CLOCK_REG);  
    break;
    
  case ADC_SET_MODE:
    if (minor > AD_CHANNELS)
      return(-EINVAL);
    Chan[minor].mode = (BYTE) arg;
    if (Chan[minor].mode == ADC_PACER_EXTERN_TRIG) {
      BoardData.pacerReg |= 0x1;
    } else {
      BoardData.pacerReg &= ~(0x1);
    }
    break;
    
  case DIO_SET_MODE:
    if (minor < AD_CHANNELS || minor > (AD_CHANNELS + DIO_PORTS))
      return(-EINVAL);
    
    arg &= 0x3;
    switch (minor & 0xf) {
    case 0:			/* Port 0 */
#ifdef DEBUG
      printk("DIO_SET_MODE for Port 0\n");
#endif
      return(-EINVAL);		/* These are split mode I/O bits */
      break;
    case 1:			/* Port A */
#ifdef DEBUG
      printk("DIO_SET_MODE for Port 1\n");
#endif
      bReg = inb(DIO_CNTRL_REG) & 0x9f;
      bReg |= (arg << 5);
      outb_p(bReg, DIO_CNTRL_REG);
      break;
    case 2:			/* Port B */
#ifdef DEBUG
      printk("DIO_SET_MODE for Port 2\n");
#endif
      if (arg & 0x2)
	return(-EINVAL);	/* Port B only has Modes 0 & 1 */
      bReg = inb(DIO_CNTRL_REG) & 0xfb;
      bReg |= (arg << 2);
      outb_p(bReg, DIO_CNTRL_REG);
      break;
    case 3:			/* Port C */
#ifdef DEBUG
      printk("DIO_SET_MODE for Port 3\n");
#endif
      if (arg)
	return(-EINVAL);	/* Port C only has Mode 0 I/O */
      break;
    default:
#ifdef DEBUG
      printk("DIO_SET_MODE for Invalid Port\n");
#endif
      return(-EINVAL);		/* Wrong Port Number */
      break;
    }
    break;
    
  case DIO_SET_DIRECTION:
    if (minor < AD_CHANNELS || minor > (AD_CHANNELS + DIO_PORTS))
      return(-EINVAL);
    
    switch (minor & 0xf) {
    case 0:			/* Port 0 */
#ifdef DEBUG
      printk("DIO_SET_DIRECTION for Port 0\n");
#endif
      return(-EINVAL);		/* You cannot change the direction of these bits */
      break;
    case 1:			/* Port A */
#ifdef DEBUG
      printk("DIO_SET_DIRECTION for Port 1\n");
#endif
      arg &= 0x1;
      bReg = inb(DIO_CNTRL_REG) & 0xef;
      bReg |= (arg << 4);
      outb_p(bReg, DIO_CNTRL_REG);
      break;
    case 2:			/* Port B */
#ifdef DEBUG
      printk("DIO_SET_DIRECTION for Port 2\n");
#endif
      arg &= 0x1;
      bReg = inb(DIO_CNTRL_REG) & 0xfd;
      bReg |= (arg << 1);
      outb_p(bReg, DIO_CNTRL_REG);
      break;
    case 3:			/* Port C */
#ifdef DEBUG
      printk("DIO_SET_DIRECTION for Port 3\n");
#endif
      arg &= 0x3;
      bReg = inb(DIO_CNTRL_REG) & 0xf6;
      bReg |= (arg << (arg&0x2?2:0));
      outb_p(bReg, DIO_CNTRL_REG);
      break;
    default:
#ifdef DEBUG
      printk("DIO_SET_DIRECTION for Invalid Port\n");
#endif
      return(-EINVAL);
      break;
    }
    break;
    
  default:
    return(-EINVAL);
    break;
  }

  return 0;
}

/***************************************************************************
 *
 *
 ***************************************************************************/

static struct file_operations das_fops = {
  NULL,				/* seek    */
  das_read,			/* read    */
  das_write,			/* write   */
  NULL,				/* readdir */
  NULL,				/* select  */
  das_ioctl,			/* ioctl   */
  NULL,				/* select  */
  das_open,			/* open    */
  das_close,			/* close   */
  NULL				/* fsync   */
};

/***************************************************************************
 *
 * Loads driver. Called when "insmod adc.o" is invoked on the command line.
 *               The board is set to IRQ, 
 *
 ***************************************************************************/

int init_module(void) 
{
  register int i, retry;
  register BYTE bReg;
#ifdef DEBUG
  register int j;
#endif
  
  retry = 0;
  do {
    DMAbuf[0] = (WORD*) kmalloc(BUFSIZ * sizeof(WORD), GFP_DMA);
  } while ((DMAbuf[0] == NULL) && (retry < RETRY_LIMIT));

  retry = 0;
  do {
    DMAbuf[1] = (WORD*) kmalloc(BUFSIZ * sizeof(WORD), GFP_DMA);
  } while ((DMAbuf[1] == NULL) && (retry < RETRY_LIMIT));

  if (!DMAbuf[0] || !DMAbuf[1]) {
    printk("%s: Unable to allocate DMA buffers of %d bytes\n",
	   ADAPTER_ID, BUFSIZ * 2 * sizeof(WORD));
    return -EIO;
  }

  /* Register as a device with kernel.  */

  if (register_chrdev(MajorNumber, ADAPTER_ID, &das_fops)) {
    printk("%s: Unable to register character device.\n", ADAPTER_ID);
    return -EIO;
  } else {
    printk("%s: BASE=%#x IRQ=%d DMA=%d DMABufsiz=%d.\n",
	   ADAPTER_ID, BASE_REG, DAS_IRQ, DAS_DMA, BUFSIZ);
    printk(" 4/27/95 wjasper@tx.ncsu.edu sam@tx.ncsu.edu\n");
    printk("10/13/95 jfbeam@tx.ncsu.edu rncox@tx.ncsu.edu\n");
  }
 
  cli();
  /* Register interrupt handler. */
  if (request_irq(DAS_IRQ, das0_ReadInterrupt, SA_INTERRUPT, ADAPTER_ID)) {
    sti();
    printk("%s: Can't request IRQ %d\n", ADAPTER_ID, DAS_IRQ);
    return -(EIO);
  }

  /* Register DMA. */
  if (request_dma(DAS_DMA, ADAPTER_ID)) {
    sti();
    printk("%s: Can't request DMA %d\n", ADAPTER_ID, DAS_DMA);
    free_irq(DAS_IRQ);
    return -(EIO);
  }

  /* Set all channel structures to show nothing active/open */
  for (i = 0; i < AD_CHANNELS; i++) {
    Chan[i].open = FALSE;
    Chan[i].adc_dio_pretrig = FALSE;
  }

#ifdef DEBUG
  for (i = 0, j = 0; i < BUFSIZ; i++) {
    DMAbuf[0][i] = (WORD) j++;
    DMAbuf[1][i] = (WORD) j++;
  }

  printk("%s: DMAbuf[0] == %p  DMAbuf[1] == %p\n",
	 ADAPTER_ID, DMAbuf[0], DMAbuf[1]);
#endif

  das0_wait = NULL;

  /* Set interrupt level on board */
  bReg = DAS_IRQ;
  bReg <<= 4;			/* Tell ADC which IRQ to use */
  outb_p(bReg, CNTRL_REG);	/* ... clearing INTE bit in control reg */

  BoardData.freq = DEFAULT_FREQ; /* Set default pacer clock frequency */
  SetPacerFreq(&BoardData);
  outb_p(0x0, MISC_REG);	/* Put in compatibility mode */

  outb_p(0x40, MODE_ENABLE_REG); /* DAS1600 Functions ENABLED */
  outb_p(0x0, BURST_MODE_REG);	/* Burst Mode DISABLED        */
  outb_p(0x0, CONVERT_DIS_REG);	/* Conversions are ENABLED   */

  bReg = inb_p(BURST_STAT_REG);
  printk("%s:     Burst Mode %s\n", ADAPTER_ID,
	 bReg&DAS1600_BME?"Enabled":"Disabled");
  printk("%s:     DAS1600 Mode %s\n", ADAPTER_ID,
	 bReg&DAS1600_ME?"Enabled":"Disabled");
  printk("%s:     Conversions %s\n", ADAPTER_ID,
	 bReg&DAS1600_CD?"Allowed":"Blocked");
  printk("%s:     WaitState %s\n", ADAPTER_ID,
	 bReg&DAS1600_WS?"Enabled":"Disabled");
  printk("%s:     Using %dMHz Clocks\n", ADAPTER_ID,
	 bReg&DAS1600_CLK?10:1);

  sti();

#ifdef DEBUG
  printk("%s: Setup complete. IRQ set to %d\n", ADAPTER_ID, DAS_IRQ);
#endif

  return 0;
}

/***************************************************************************
 *
 * Remove driver. Called when "rmmod adc" is run on the command line.
 *
 ***************************************************************************/

void cleanup_module(void) 
{
  if (MOD_IN_USE) {
    printk("%s: device busy, remove delayed\n", ADAPTER_ID);
    return;
  }

  /* Release the DMA channel and the IRQ */
  free_dma(DAS_DMA);
  free_irq(DAS_IRQ);

  /* Return the DMA buffers to free memery */
  kfree(DMAbuf[0]);
  kfree(DMAbuf[1]);

  if (unregister_chrdev(MajorNumber, ADAPTER_ID) != 0) {
    printk("%s: Unable to unregister character device.\n", ADAPTER_ID);
  } else {
    printk("%s: module removed.\n", ADAPTER_ID);
  }    
}

/***************************************************************************
 *
 * Block copy of data samples from data register to kernel memory.
 *
 ***************************************************************************/

inline char *RegCopy(char *dest, const int reg, int n)
{
  __asm__ __volatile__(
		       "cld\n\t"
		       "rep\n\t"
		       "insw"
		       : 
		       :"D" (dest), "d" (reg),"c" (n)
		       :"di","dx","cx");
  return dest;
}

/*
 * DMA transfer routine
 */

static void doDMA(int samples)
{
  BYTE bReg;
  unsigned long flags;

  if(samples) {
    /* Ok, we have been asked to grab a block of data with the DMA */

    save_flags(flags);
    cli();

    /* The DMA controller has to be setup before the DAS-1600 is told
       to use DMA transfer mode.  So, setup the DMA first.  DMA controller
       must be off before it can be setup.                                 */

#ifdef DEBUG
    printk("doDMA(1): DMAbuf[%d] = %p, Count = %d\n",
	   DMAptr, DMAbuf[DMAptr], samples * sizeof(WORD));
#endif

    disable_dma(DAS_DMA);
    clear_dma_ff(DAS_DMA);
    set_dma_mode(DAS_DMA, DMA_MODE_READ);
    set_dma_addr(DAS_DMA, (long) DMAbuf[DMAptr]);
    set_dma_count(DAS_DMA, samples * sizeof(WORD));
    enable_dma(DAS_DMA);

#ifdef DEBUG
    printk("doDMA(2): DMAbuf[%d] = %p, Count = %d\n",
	   DMAptr, DMAbuf[DMAptr], samples * sizeof(WORD));
#endif

    /* With the DMA controller now setup and active, we can tell the card
       to use the DMA transfer mode.                                      */

    bReg = inb_p(CNTRL_REG);
    bReg |= (CNTL_DMA | CNTL_INTE);
    outb_p(bReg, CNTRL_REG);

    outb_p(0x0, STATUS_REG);

#ifdef DEBUG
    printk("doDMA(): DMA triggered and INT is enabled\n");
#endif

    restore_flags(flags);
  } else {
    /* Zero words passed to this routine is used to shutdown the DMA
       controller for the card.                                           */

    disable_dma(DAS_DMA);

    bReg = inb_p(CNTRL_REG);
    bReg &= ~(CNTL_INTE | CNTL_DMA);
    outb_p(bReg, CNTRL_REG);

#ifdef DEBUG
    printk("doDMA(): DMA & INT are disabled\n");
#endif
  }
  return;
}

/***************************************************************************
 *
 *  Alarm handler to handle situations where an interrupt is missed.
 *
 ***************************************************************************/

void das0_TimerHandler(LONG unused)
{
  /* we should only be here if an interrupt
     occured while we were reading data      */

#ifdef DEBUG
  printk("TimerHandler: WordsToRead = %d\n", WordsToRead);
#endif

  /* This is bad, so stop everything and alert the user the data is corrupt */

  WordsToRead = -1;
  ISR_WordsToRead = -1;
  doDMA(0);
  wake_up_interruptible(&das0_wait);  
  return;
}

/***************************************************************************
 *
 * Interrupt handler used to service enhanced mode(interrupt) read().
 *
 ***************************************************************************/

void das0_ReadInterrupt(int irq)
{
  register BYTE msb, lsb;
  WORD sample;

  cli();
  switch (CurrChan->mode) {
  case ADC_SOFT_TRIGGER:
    /* Bang! Bang! Method: Force conversion, Conversion forces interrupt */
    lsb = inb(BASE_REG);	/* Sample: ADC -> user space */
    msb = inb(MSB_DATA_BYTE);
    sample = (msb << 8) | lsb;

    /* Ok, so this could be a problem... */
    put_fs_word(sample, (WORD*) KernBuff); /* store the sample          */
    WordsToRead--;		/* One sample at a time      */
    KernBuff += 1;

    if (WordsToRead) {
      outb_p(0x0, STATUS_REG);	/* Allow Interrupt      */
      outb_p(0x1, BASE_REG);	/* Force conversion     */
      sti();
    } else {
      wake_up_interruptible(&das0_wait);  
    }
    break;

  case ADC_PACER_CLOCK:
  case ADC_EXTERNAL_TRIGGER:
  case ADC_PACER_EXTERN_TRIG:
#ifdef DEBUG
    printk("das0_ReadInterrupt(DMA): WordsToRead = %d  KernBuff = %p\n",
	   WordsToRead, KernBuff);
#endif

    if (WordsToRead > BUFSIZ) {
#ifdef DEBUG
      printk("das0_ReadInterrupt(DMA): More to Read\n");
#endif

      WordsToRead -= BUFSIZ;
      DMAptr ^= 1;

      del_timer(&TimerList);
      if (WordsToRead > BUFSIZ) {
	doDMA(BUFSIZ);
	TimerList.expires = BUFSIZ * HZ / BoardData.freq + 2;
      } else {
	doDMA(WordsToRead);
	TimerList.expires = (WordsToRead) * HZ / BoardData.freq + 2;
      }

#ifdef DEBUG
      printk("%s: Timer Expires in %d\n", ADAPTER_ID, (int) TimerList.expires);
#endif

      add_timer(&TimerList);

      wake_up_interruptible(&das0_wait); 
    } else if (WordsToRead > 0) {
#ifdef DEBUG
      printk("das0_ReadInterrupt(DMA): Finished\n");
#endif

      del_timer(&TimerList);
      doDMA(0);

      WordsToRead = 0;
      
      wake_up_interruptible(&das0_wait);
    }
    break;

  default:
    break;
  }
}

/***************************************************************************
 *
 * Handles software triggered read().
 *
 * Bang! Bang! Method: 
 *    
 *    o Runs in Compatibility Mode
 *    o Force interrupt by forcing conversion.
 *    o Get one sample at a time and put it in the kernel buffer(kernBuff).
 *    o Get chan->count samples
 *
 ***************************************************************************/

static int SoftTrigRead(ChanRec *chan)
{
  BYTE bReg;

  /* Set interrupt level on board, disable interrupts, and pacer  */

  bReg = DAS_IRQ;

  bReg <<= 4;			/* Tell ADC which IRQ to use       */
  outb_p(bReg, CNTRL_REG);	/* & clear INTE bit in control reg */

  bReg = (BYTE)inb_p(MUX_SCAN_LIMITS); /* Read/Write clears FIFO          */
  outb_p(bReg, MUX_SCAN_LIMITS);	

  bReg = inb_p(CNTRL_REG);	/* Turns on ADC interrupts          */
  bReg |= CNTL_INTE;
  outb_p(bReg, CNTRL_REG);

  outb_p(0x00, STATUS_REG);	/* Clears INT bit, allow Int.       */

  /* Prepare global data for parameterless interrupt handler              */

  CurrChan = chan;		/* pass chanel number to global     */
  WordsToRead = chan->count;

  outb_p(0x1, BASE_REG);	/* Force first conversion           */
  interruptible_sleep_on(&das0_wait); /* Block in wait state              */

  /* turn off interrupts */

  bReg = inb(CNTRL_REG);
  outb_p(bReg & ~CNTL_INTE, CNTRL_REG);
  return 0;
}

/***************************************************************************
 *
 * Set voltage gains.
 *
 ***************************************************************************/

static void SetGains( BYTE gain )
{
  BYTE mask;
  BYTE test;
  
  mask = inb_p(MISC_REG);	/* read the Analog Input Range Register */
  mask &= 0xf0;			/* mask off last 4 bits                 */
  mask |= gain;			/* set gain                             */
  outb_p(mask, MISC_REG);
  test = inb_p(MISC_REG);
#ifdef DEBUG
  printk("%s:     Gain Setting:  %#x \n", ADAPTER_ID, gain); 
#endif
  return;
}

/****************************************************************************
 *   Set Pacer Clock Frequency
 * 
 *   Description: 
 *       Set the counters so that the pacer clock runs at the
 *       desired frequency.  The frequency is generated by dividing
 *       down a 10 MHz clock, so all frequencies can not be generated.
 *       This routine calculated the divisor to generate a frequency
 *       as near as possible to the requested one.  It then calculates
 *       the real frequency and returns it .
 ****************************************************************************/

static int SetPacerFreq(BoardRec *board)
{
  short unsigned ctr1, ctr2;
  LONG product, error;

  /* divide 10Mhz by frequency */

  /* This assumes the board is set for 10Mhz -- RB */

  product =  10000000.0 / board->freq + 0.5;

  /*
     Now the job is to find two 16 bit numbers, that when multiplied
     together are approximately equal to product.  Start by setting
     one of them, ctr1 to 2 (minimum settable value) and increment until
     the error is minimized and ctr2 is less than 32768.
     
     NOTE: In Mode 2, a value of 1 is illegal! Therefore, crt1 and crt2
     can never be 1.
     */

  ctr1 = product / 32768;
  if ( ctr1 < 2 ) ctr1 = 2;
  ctr2 = product / ctr1;
  error = abs(product - ctr2*ctr1);

  while ( error && ctr1 < 32768  && ctr2 > 1 ) {
    ctr1++;
    ctr2 = product / ctr1;
    error = abs(product - ctr2*ctr1);
  }

  /* the frequency is prime, add 1 to it */

  if ( error ) {
    product++;
    ctr1 = product / 32768;
    if ( ctr1 < 2 ) ctr1 = 2;
    ctr2 = product / ctr1;
    error = abs(product - ctr2*ctr1);

    while ( error  && ctr1 < 32768 && ctr2 > 1) {
      ctr1++;
      ctr2 = product / ctr1;
      error = abs(product - ctr2*ctr1);
    }
  }

  /* we can't have ctr2 equal to 1, or system hangs */

  if ( ctr2 == 1 ) {
    ctr2++;
    ctr1 /= 2;
  }

  board->ctr1 = ctr1;
  board->ctr2 = ctr2;
  board->freq = 10000000.0/(ctr1*ctr2) + 0.5;

#ifdef DEBUG
  printk("SetPacerFreq: Pacer Register set to %#x\n", BoardData.pacerReg);
#endif

  return 0;
}

/***************************************************************************
 *
 * Load two part frequency to pacer counter chip.
 *
 ***************************************************************************/

static void LoadPacer(BoardRec *board)
{
  BYTE mask;
  BYTE bData;

  /* Write the values of ctr1 and ctr2 into counter A1 and A2 */

  cli();

#ifdef DEBUG
  printk("LoadPacer: load values: A ctr1 %#x ctr2 %#x\n",
	 board->ctr1, board->ctr2);
#endif

  mask = C2 + MODE2 + LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL); 

  bData = (BYTE) (board->ctr2 & 0xff);
  outb_p(bData, COUNTERA_2_DATA); 

  bData = (BYTE) (board->ctr2 >> 8);
  outb_p(bData, COUNTERA_2_DATA); 

  mask = C1 + MODE2 + LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL); 

  bData = (board->ctr1 & 0xff);
  outb_p(bData, COUNTERA_1_DATA); 

  bData = (board->ctr1 >> 8) & 0xff;
  outb_p(bData, COUNTERA_1_DATA); 

  sti();
}

/***************************************************************************
 *
 * Load value into Counter 0  XXXX    Mode    MSB     LSB
 *                            Byte 3  Byte 2  Byte 1  Byte 0
 *
 ***************************************************************************/

static void LoadCounter0( LONG value )
{
  BYTE mask;
  BYTE bData;

  cli();

  /* Write the value into Counter 0 Mode 2 */
  
#ifdef DEBUG
  printk("LoadCounter0: load value %#lx into Counter 0.\n", value);
#endif

  /* the mode is in the thrid byte */

  mask = 0xff & (value >> 16);
  mask += C0 + LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL); 

  if ( value & 0xff000000 ) {
    /* load control word only */
    return;
  } else {
    /* LSB in the first byte of value */

    bData = (BYTE) (value & 0xff);
    outb_p(bData, COUNTERA_0_DATA); 

    /* MSB in the second byte of value */

    bData = (BYTE) ((value >> 8) & 0xff);
    outb_p(bData, COUNTERA_0_DATA); 
  }
  sti();
}

/***************************************************************************
 *
 * Set which channels read() is interested in.
 *
 ***************************************************************************/

static void SetChannelMux(BYTE lowChan, BYTE highChan)
{
  BYTE channelMask;

  channelMask = (highChan << 4) | lowChan;
  outb_p(channelMask, MUX_SCAN_LIMITS);
  
#ifdef DEBUG
  printk("%s: SetChannelMux: lowChan = %d  highChan = %d\n",
	 ADAPTER_ID, lowChan, highChan);
#endif
  return;
}

/***************************************************************************
 *
 * Turn off pacer timer chip.
 *
 ***************************************************************************/

static void StopPacer()
{
  BYTE mask;

  cli();

  mask = C2 + MODE2 + LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL);

  mask = C1 + MODE2 + LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL);
  sti();
}

/***************************************************************************
 *
 * Handles 
 *    o pacer/counter triggered read() with software start.
 *    o pacer/counter triggered read() with external start.
 *    o external triggered read().
 *
 * Pacer Method: 
 *
 *    o Runs in CIO-DAS 16/330 Enhanced Mode
 *    o Pacer counter controls frequency of sample conversions
 *    o Total counter controls number of samples taken
 *    o Interrupts occur every 512(PACKETSIZE) samples and at last sample
 *    o Get chan->count samples
 *    o Samples are saved in kernel buffer(kernBuff) in readInterrupt()
 *
 ***************************************************************************/

static int PacerRead(ChanRec *chan)
{
  BYTE bReg = 0x0;

#ifdef DEBUG
  printk("Entering PacerRead().\n");
#endif

  outb_p(0x40, BURST_MODE_REG);	/* Turns Burst Mode On!!  */
  bReg = inb_p(BURST_STAT_REG);
#ifdef DEBUG
  printk("%s:     Burst Mode %s\n", ADAPTER_ID, bReg&DAS1600_BME?
	 "Enabled":"Disabled");
#endif

  /* Set interrupt level on board, disable interrupts, and pacer  */
  bReg = DAS_IRQ;
  bReg <<= 4;			/* Tell ADC which IRQ to use       */
  outb_p(bReg, CNTRL_REG);	/* & clear INTE bit in control reg */

  bReg = (BYTE)inb_p(MUX_SCAN_LIMITS); /* Read/Write clears FIFO          */
  outb_p(bReg, MUX_SCAN_LIMITS);	
#ifdef DEBUG
  printk("%s:     Mux Register Setting(1) = %#x\n", ADAPTER_ID, bReg);
#endif

  SetGains(chan->gain);

  bReg = (BYTE)inb_p(CNTRL_REG); /* S1=0 and S0=0 from above          */
  if((chan->mode == ADC_PACER_CLOCK) ||
     (chan->mode == ADC_PACER_EXTERN_TRIG)) {
    bReg |= (0x3);		/* Set S1=1 and S0=1 Internal Pacer  */
  } else {
    bReg |= (0x2);		/* Set S1=1 and S0=0 External Pacer  */
    BoardData.freq = 1;         /* Remove the timer from the eq.     */
  }
  outb_p(bReg, CNTRL_REG);

  if( chan->mode == ADC_PACER_CLOCK || chan->mode == ADC_PACER_EXTERN_TRIG ) {
    LoadPacer(&BoardData);	/* Establish sample frequency        */
  }

  /* Prepare global data for parameterless interrupt handler                */

  CurrChan = chan;		/* pass chanel number to global      */
  ISR_WordsToRead = chan->count;
  WordsToRead = chan->count;
  
  if( chan->mode == ADC_PACER_CLOCK || chan->mode == ADC_PACER_EXTERN_TRIG ) {

    /* Start pacer.  Conversions start when CONVERT_DIS_REG is cleared */

    outb_p(BoardData.pacerReg, PACER_CLOCK_REG);
  }

  if ( chan->adc_dio_pretrig ) {
    outb_p( chan->dio_value, DIO_REG );
  }

  cli();

  outb_p(0x0, STATUS_REG);	/* Clear INT bit, allow Interrupt    */

  bReg = (BYTE)inb_p(MUX_SCAN_LIMITS); /* Read/Write clears FIFO            */
  outb_p(bReg, MUX_SCAN_LIMITS);
#ifdef DEBUG
  printk("%s:     Mux Register Setting(2) = %#x\n", ADAPTER_ID, bReg);
#endif

  outb_p(0x0, STATUS_REG);	/* Allow Interrupt                   */

  outb_p(0x0, CONVERT_DIS_REG);	/* Enable Conversions to the A/D     */

  /* Activate the DMA controller */

  DMAptr = 0;
  del_timer(&TimerList);
  if (WordsToRead > BUFSIZ) {
    doDMA(BUFSIZ);
    TimerList.expires = BUFSIZ * HZ / BoardData.freq + 2;
  } else {
    doDMA(WordsToRead);
    TimerList.expires = (WordsToRead) * HZ / BoardData.freq + 2;
  }
  sti();

#ifdef DEBUG
  printk("%s: Timer Expires in %d\n", ADAPTER_ID, (int) TimerList.expires);
#endif
  
  add_timer(&TimerList);

#ifdef DEBUG
  bReg = (BYTE)inb_p(CNTRL_REG);
  printk("PacerRead(): CNTRL_REG = %#x\n", bReg);
#endif

  do {
    /* Wait for the First (of 1+) blocks to be collected */

#ifdef DEBUG
    printk("PacerRead(): Waiting for block\n   WordsToRead = %d\n",
	   ISR_WordsToRead);
#endif

    interruptible_sleep_on(&das0_wait);
    if (ISR_WordsToRead > BUFSIZ) {

      /* The DMA/INT have transfered one BUFSIZ buffer full of data */

      /* The DMA transfer is active on DMAptr, the other DMAbuf
	 is the Buffer with the data in it.                       */

#ifdef DEBUG
      printk("PacerRead(): Awake for block\n");
#endif

      cli();
      memcpy_tofs(KernBuff, DMAbuf[DMAptr ^ 1], BUFSIZ * sizeof(WORD));
      sti();
      KernBuff += BUFSIZ;
      ISR_WordsToRead -= BUFSIZ;
     
    } else if (ISR_WordsToRead > 0){

      /* In this case, no DMA transfer is active, thus the DMAptr
	 is pointing the the data.                                */

#ifdef DEBUG
      printk("PacerRead(): Awake for LAST block\n");
#endif

      cli();
      memcpy_tofs(KernBuff, DMAbuf[DMAptr],
		  ISR_WordsToRead * sizeof(WORD));
      sti();
      KernBuff = NULL;
      ISR_WordsToRead = 0;

    }
  } while (ISR_WordsToRead > 0);

  /* turn off interrupts, and trigger control.
     Leave pacer clock on if enabled          */

  outb_p(0x0, CNTRL_REG);

  if ( WordsToRead != 0 ) {
    printk("Timing error in PacerRead: WordsToRead = %d\n", WordsToRead);
    return -1;
  }

  outb_p(0x00, STATUS_REG);	/* Clear INT bit */
  outb_p(0x00, BURST_MODE_REG);	/* Turns Burst Mode Off */
  
  bReg = inb_p(BURST_STAT_REG);
#ifdef DEBUG
  printk("%s:     Burst Mode %s\n", ADAPTER_ID, bReg&DAS1600_BME?
	 "Enabled":"Disabled");
#endif
  return 0;
}
